--
-- mpm_apa.vhd
--
-- This file implements a variable width "Quad-port" memory or a 
-- "pseudo-true dual port" memory (upto 72 bits wide) based on the
-- 256x9 memory blocks avaialble in Actel's APA Family of FPGAs. 
-- The user must gang these blocks together or modify this wrapper
-- to accomodate deeper memories.
--
-- The memory implemented in this file offers 4 access ports: 2 write
-- ports and 2 read ports (QPM mode) or 2 write/read ports (DPM mode).
--
-- Two clocks - one full rate clock (WR_2xCLK) and one half-rate clock
-- (WR_CLK) are shared by the write ports.
--
-- Similarly one half-rate clock (RD_CLK) is shared by the read ports.
--
-- Write ports:
--  Data, address, and enables for both write ports are optionally 
--  registered with the rising edge of WR_CLK when PIPE = 1.
--  Then during the high portion of WR_CLK, the values for WPa are muxed 
--  to the memory blocks and settle, then data is written into the memory
--  on the rising edge of WR_2xCLK (ie the falling edge of WR_CLK).
--  During the low portion of WR_CLK, the values for WPb are muxed to the
--  memory and allowed to settle.  Then on the rising-edge of WR_2xCLK
--  (ie: the next rising-edge of WR_CLK), this value is written into
--  the memory while the next write-port values are captured.
--
-- READ ports (DPM mode:PMODE = 0)
--
-- When PMODE is set to '0' (DPM), the read address inputs and read clock are
-- unused and the write addresses are used to read the memory.  IF OREG = '0',
-- the data outputs propagate directly to the rd_out ports; otherwise, when
-- OREG = '1', RDa is registered on the falling-edge of WR_CLK, and RDb is
-- registered on the rising-edge of WR_CLK.  The read enable inputs are used
-- as inputs to enable the output regsters.
--
-- NOTE: the APA RAM256X9SA is used as the basic memory for this wrapper, so
-- the write address propagates through the read port and sets-up data to the
-- output ports/registers BEFORE the write operation is performed.  So this is
-- a READ, then WRITE architecture.  A simultaneous read/write cycle will 
-- result in reading the OLD data followed by the data begin updated in the 
-- memory at the end of the clock cycle.  The next read operation to that 
-- address will caputure the value just written.
--
-- READ ports (QPM mode:PMODE = 1)
--   
--  If PIPE is set to '1', the read address and enables for both read
--  ports are registered with the rising edge of the read clock (RD_CLK).
--  Then during the high portion of RD_CLK, the value for RDA is muxed to
--  the memory block and RDa will propagate through the memory block.
--  Then, if OREG is set to '0' the result is wired directly to RDa 
--  (transparent outputs), otherwise, if OREG is set to '1', a register will
--  capture the result on the falling-edge of RD_CLK and then forward it to
--  the output re-synchronized with the rising-edge of rd_clk.  Then during
--  the low portion of RD_CLK, the address for RDb is muxed to the memory blocks
--  and RDb propagates from the memory.  Again, if OREG = '0', then the result
--  is wired directly to RDb, otherwise an register will capture RDb on the
--  rising-edge of RD_CLK.
--
--  If PIPE is set to '0', the read address and enables are not
--  registered and only the mux is provided for Raddr_a and Raddr_b.
--
-- PARAMTER    VALUE    DESCRIPTION
--
--  PMODE        0      Dual-port operation      
--               1      Four-port operation  (Default)
--
--  PIPE         0      No input registers, only mux logic for inputs (Default)
--               1      Register inputs, then muliplex inputs to the memory
--
--  OREG         0      Transparent output mode  (default)
--               1      Registered  output mode
--
--  WIDTH       1:72    Number of data + parity bits (default = 9)
--
--
--  Here's a sample instantiation of the block (cut and paste into
--  your higher-level VHDL code):
--
--  QPM0: mpm_apa
--  GENERIC MAP (PMODE => 0, PIPE = 0, OREG = 0, WIDTH = 9);
--
--  PORT MAP    (reset_n  => <your reset>,
--               wr_2xclk => <2x write clock>,
--               wr_clk   => <write port data/enable clock>,
--
--               a_wrblk  => <active low block enable for write>,
--               a_wrb    => <active low data enable for write>,
--               a_wadr   => <write/dpm address bus  (8-bits)>,
--               a_wdata  => <write data bus (variable width)>,
--   
--               b_wrblk  => <active low block enable for write>,
--               b_wrb    => <active low data enable for write>,
--               b_wadr   => <write/dpm address bus  (8-bits)>,
--               b_wdata  => <write data bus (variable width)>,
--
--               rd_clk   => <read data clock (QPM mode)>,
--
--               a_rdblk  => <output register enable 1>,
--               a_rdb    => <output register enable 2>,
--               a_radr   => <output address bus 8-bits (QPM mode)>,
--               a_rdata  => <output data bus (variable width)>,
--
--               b_rdblk  => <output register enable 1>,
--               b_rdb    => <output register enable 2>,
--               b_radr   => <output address bus 8-bits (QPM mode)>,
--               b_rdata  => <output data bus (variable width)>);
--
------------------------------------------------------------------------
--
--                  Copyright 2002 Actel corporation
--
------------------------------------------------------------------------
--
-- Version 1.0  11/08/02 J.Vorgert - Initial release
-- Version 1.1  03/09/03 J.Vorgert - Fixed typo in PMODE=0;OREG=1 that
--                                   caused incorrect read enable logic
-- Version 1.2  04/12/03 J.Vorgert - No changes to this file
--
------------------------------------------------------------------------
--

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;

-- depending on the tools suite used, the APA library
-- may need to be referenced for VITAL simulation models.

--library apa;  
--use apa.all;

-- NOTE: Integer types are used for all GENERIC declarations 
-- in order to include synopsys support.

entity mpm_apa is
    generic (PMODE      : INTEGER range 0 to 1  := 1;   -- QPM
             PIPE       : INTEGER range 0 to 1  := 1;   -- M, R
             OREG       : INTEGER range 0 to 1  := 1;   -- M, R
             WIDTH      : INTEGER range 1 to 72 := 36); -- 1:72
 
    port( reset_n   : in  std_logic;  -- active low
          
          wr_2xclk  : in  std_logic;
          wr_clk    : in  std_logic;

          a_wrblkb  : in  std_logic; 
          a_wrb     : in  std_logic; 
          a_wadr    : in  std_logic_vector(7 downto 0);
          a_wdata   : in  std_logic_vector(WIDTH - 1 downto 0);

          b_wrblkb  : in  std_logic; 
          b_wrb     : in  std_logic; 
          b_wadr    : in  std_logic_vector(7 downto 0);
          b_wdata   : in  std_logic_vector(WIDTH - 1 downto 0);

          rd_clk    : in  std_logic;

          a_rdblkb  : in  std_logic; 
          a_rdb     : in  std_logic; 
          a_radr    : in  std_logic_vector(7 downto 0);
          a_rdata   : out std_logic_vector(WIDTH - 1 downto 0);

          b_rdblkb  : in  std_logic; 
          b_rdb     : in  std_logic; 
          b_radr    : in  std_logic_vector(7 downto 0);
          b_rdata   : out std_logic_vector(WIDTH - 1 downto 0));

end mpm_apa;

architecture RTL of mpm_apa is

-- APA 256x9 RAM with parity checking (parity not used)

component RAM256x9SA
port(
DO8 		:out 	STD_ULOGIC;
DO7 		:out 	STD_ULOGIC;
DO6 		:out 	STD_ULOGIC;
DO5 		:out 	STD_ULOGIC;
DO4 		:out 	STD_ULOGIC;
DO3 		:out 	STD_ULOGIC;
DO2 		:out 	STD_ULOGIC;
DO1 		:out 	STD_ULOGIC;
DO0 		:out 	STD_ULOGIC;
WPE 		:out 	STD_ULOGIC;
RPE 		:out 	STD_ULOGIC;
DOS 		:out 	STD_ULOGIC;
WADDR7 	:in 	STD_ULOGIC;
WADDR6 	:in 	STD_ULOGIC;
WADDR5 	:in 	STD_ULOGIC;
WADDR4 	:in 	STD_ULOGIC;
WADDR3 	:in 	STD_ULOGIC;
WADDR2 	:in 	STD_ULOGIC;
WADDR1 	:in 	STD_ULOGIC;
WADDR0 	:in 	STD_ULOGIC;
RADDR7 	:in 	STD_ULOGIC;
RADDR6 	:in 	STD_ULOGIC;
RADDR5 	:in 	STD_ULOGIC;
RADDR4 	:in 	STD_ULOGIC;
RADDR3 	:in 	STD_ULOGIC;
RADDR2 	:in 	STD_ULOGIC;
RADDR1 	:in 	STD_ULOGIC;
RADDR0 	:in 	STD_ULOGIC;
WCLKS 	:in 	STD_ULOGIC;
DI8 		:in 	STD_ULOGIC;
DI7 		:in 	STD_ULOGIC;
DI6 		:in 	STD_ULOGIC;
DI5 		:in 	STD_ULOGIC;
DI4 		:in 	STD_ULOGIC;
DI3 		:in 	STD_ULOGIC;
DI2 		:in 	STD_ULOGIC;
DI1 		:in 	STD_ULOGIC;
DI0 		:in 	STD_ULOGIC;
WRB 		:in 	STD_ULOGIC;
RDB 		:in 	STD_ULOGIC;
WBLKB 	:in 	STD_ULOGIC;
RBLKB 	:in 	STD_ULOGIC;
PARODD 	:in 	STD_ULOGIC;
DIS 		:in 	STD_ULOGIC);
end component;

SIGNAL wrblk_a  : std_logic;
SIGNAL wrb_a    : std_logic;
SIGNAL wadr_a   : std_logic_vector(7 downto 0);
SIGNAL wdata_a  : std_logic_vector(WIDTH - 1 downto 0);
SIGNAL wrblk_b  : std_logic;
SIGNAL wrb_b    : std_logic;
SIGNAL wadr_b   : std_logic_vector(7 downto 0);
SIGNAL wdata_b  : std_logic_vector(WIDTH - 1 downto 0);
SIGNAL wrblk    : std_logic;
SIGNAL wrb      : std_logic;
SIGNAL wadr     : std_logic_vector(7  downto 0);
SIGNAL wdata    : std_logic_vector(80 downto 0);

SIGNAL rdblkb_a : std_logic;
SIGNAL rdblkb_b : std_logic;
SIGNAL rdb_a    : std_logic;
SIGNAL rdb_b    : std_logic;
SIGNAL rdblkb   : std_logic;
SIGNAL rdb      : std_logic;

SIGNAL radr_a   : std_logic_vector(7 downto 0);
SIGNAL radr_b   : std_logic_vector(7 downto 0);
SIGNAL radr     : std_logic_vector(7  downto 0);
SIGNAL rd_data  : std_logic_vector(80 downto 0);
SIGNAL a_idata  : std_logic_vector(WIDTH - 1 downto 0);

SIGNAL GND      : std_logic;

begin

GND <= '0'; -- used to tie off un-used inputs to memories

--
-- Regster incoming data from the write ports
--

A: if (PIPE = 1) generate
  B: process(reset_n, wr_clk)
     begin
     if (reset_n = '0') then
 
      wrblk_a  <= '0';
      wrb_a    <= '0';
      wadr_a   <= (OTHERS => '0');
      wdata_a  <= (OTHERS => '0');

      wrblk_b  <= '0';
      wrb_b    <= '0';
      wadr_b   <= (OTHERS => '0');
      wdata_b  <= (OTHERS => '0');

     elsif (wr_clk'event and wr_clk = '1') then

      wrblk_a  <= a_wrblkb  after 1 ns;
      wrb_a    <= a_wrb     after 1 ns;
      wadr_a   <= a_wadr    after 1 ns;
      wdata_a  <= a_wdata   after 1 ns;

      wrblk_b  <= b_wrblkb  after 1 ns;
      wrb_b    <= b_wrb     after 1 ns;
      wadr_b   <= b_wadr    after 1 ns;
      wdata_b  <= b_wdata   after 1 ns;

     end if;
   end process;
end generate;

-- 
-- Otherwise, just pass 'em through to the mux
--

C: if (PIPE /= 1) generate
      wrblk_a  <= a_wrblkb;
      wrb_a    <= a_wrb;
      wadr_a   <= a_wadr;
      wdata_a  <= a_wdata;

      wrblk_b  <= b_wrblkb;
      wrb_b    <= b_wrb;
      wadr_b   <= b_wadr;
      wdata_b  <= b_wdata;
end generate;

--
-- Multiplex the write ports to the memory
--

wrblk <= wrblk_a  when (wr_clk = '1') else wrblk_b;
wrb   <= wrb_a    when (wr_clk = '1') else wrb_b;
wadr  <= wadr_a   when (wr_clk = '1') else wadr_b;

wdata(80 downto WIDTH) <= (OTHERS => '0'); -- tie off unused bits
wdata(WIDTH - 1 downto 0) <= wdata_a  when (wr_clk = '1') else
                             wdata_b(WIDTH - 1 downto 0);

--
-- IF FOUR-Port and PIPE = 1 Regster the read
-- addresses and enables
--

D: if (PMODE = 1 AND PIPE = 1) generate
   E: process(reset_n, rd_clk)
   begin
    if (reset_n = '0') then
 
      rdblkb_a <= '0';
      rdb_a    <= '0';
      radr_a   <= (OTHERS => '0');

      rdblkb_b <= '0';
      rdb_b    <= '0';
      radr_b   <= (OTHERS => '0');
 
    elsif (rd_clk'event and rd_clk = '1') then

      rdblkb_a <= a_rdblkb  after 1 ns;
      rdb_a    <= a_rdb     after 1 ns;
      radr_a   <= a_radr    after 1 ns;
 
      rdblkb_b <= b_rdblkb  after 1 ns;
      rdb_b    <= b_rdb     after 1 ns;
      radr_b   <= b_radr    after 1 ns;
 
    end if;
   end process;
end generate;

--
-- Otherwise, it's just a pass them through
--

F: if (PIPE /= 1) generate
      rdblkb_a <= a_rdblkb;
      rdb_a    <= a_rdb;
      radr_a   <= a_radr;
 
      rdblkb_b <= b_rdblkb;
      rdb_b    <= b_rdb;
      radr_b   <= b_radr;
end generate;

--
-- In four-port mode, multiplex the read addresses
-- NOTE: enables are used for output registers only.
--

W: if (PMODE = 1) generate
       radr <= radr_a when (rd_clk = '1') else radr_b;
end generate;

--
-- IF OREG is 1 (registered), then create the output
-- registers, use rd_clk in QPM (PMODE = 1) mode...
--

rd_data(80 downto WIDTH) <= (OTHERS => '0'); -- tie off unused bits.

G: if (PMODE = 1 AND OREG = 1) generate

    I: process(reset_n, rd_clk)
    begin
     if (reset_n = '0') then
        a_idata <= (OTHERS => '0');
     elsif (rd_clk'event and rd_clk = '0') then
        if(rdblkb_a = '0' AND rdb_a = '0') then
           a_idata <= rd_data(WIDTH-1 downto 0) after 1 ns;
        end if;
     end if;
    end process;

    J: process(reset_n, rd_clk)
    begin
     if (reset_n = '0') then
        b_rdata <= (OTHERS => '0');
        a_rdata <= (OTHERS => '0');
     elsif (rd_clk'event and rd_clk = '1') then
        if(rdblkb_b = '0' AND rdb_b = '0') then
           b_rdata <= rd_data(WIDTH-1 downto 0) after 1 ns;
        end if;
        if(rdblkb_a = '0' AND rdb_a = '0') then
           a_rdata <= a_idata after 1 ns;
        end if;
     end if;
    end process;

end generate;

--
--  In dual port mode - use the write clock for output registers.
--

K: if (PMODE = 0 AND OREG = 1) generate

    M: process(reset_n, wr_clk)
    begin
     if (reset_n = '0') then
        a_idata <= (OTHERS => '0');
     elsif (wr_clk'event and wr_clk = '0') then
        if(a_rdblkb = '0' AND a_rdb = '0') then
           a_idata <= rd_data(WIDTH-1 downto 0) after 1 ns;
        end if;
     end if;
    end process;

    N: process(reset_n, wr_clk)
    begin
     if (reset_n = '0') then
        b_rdata <= (OTHERS => '0');
        a_rdata <= (OTHERS => '0');
     elsif (wr_clk'event and wr_clk = '1') then
        if(b_rdblkb = '0' AND b_rdb = '0') then
           b_rdata <= rd_data(WIDTH-1 downto 0) after 1 ns;
        end if;
        if(a_rdblkb = '0' AND a_rdb = '0') then
           a_rdata <= a_idata after 1 ns;
        end if;
     end if;
    end process;

end generate;

--
-- otherwise, assign the ouput of the 
-- memory directly to the read data ports.
--

O: if (OREG /= 1) generate
    a_rdata <= rd_data(WIDTH-1 downto 0); -- assign to output ports
    b_rdata <= rd_data(WIDTH-1 downto 0); 
end generate;

--
-- IF QPM generate (WIDTH/9) RAM BLOCKS with separate read/write ports
--

Q: if (PMODE = 1) generate
   R: for i in 0 to (WIDTH/9) generate       
      S: RAM256x9SA
         port map( DI8 => wdata((i*9)+8),  DI7 => wdata((i*9)+7),
                   DI6 => wdata((i*9)+6),  DI5 => wdata((i*9)+5),
                   DI4 => wdata((i*9)+4),  DI3 => wdata((i*9)+3),
                   DI2 => wdata((i*9)+2),  DI1 => wdata((i*9)+1),
                   DI0 => wdata((i*9)+0),

                   WADDR7 => wadr(7),  WADDR6 => wadr(6),
                   WADDR5 => wadr(5),  WADDR4 => wadr(4),
                   WADDR3 => wadr(3),  WADDR2 => wadr(2),
                   WADDR1 => wadr(1),  WADDR0 => wadr(0),

                   WBLKB  => wrblk,    WRB => wrb,
                   WCLKS  => wr_2xclk,
                   
                   WPE => open, RPE => open, DOS => open,

                   RBLKB  => GND,   RDB => GND,

                   RADDR7 => radr(7), RADDR6 => radr(6),
                   RADDR5 => radr(5), RADDR4 => radr(4),
                   RADDR3 => radr(3), RADDR2 => radr(2),
                   RADDR1 => radr(1), RADDR0 => radr(0),

                   DO8 => rd_data((i*9)+8), DO7 => rd_data((i*9)+7),
                   DO6 => rd_data((i*9)+6), DO5 => rd_data((i*9)+5),
                   DO4 => rd_data((i*9)+4), DO3 => rd_data((i*9)+3),
                   DO2 => rd_data((i*9)+2), DO1 => rd_data((i*9)+1),
                   DO0 => rd_data((i*9)+0),

                   PARODD => GND, DIS => GND
         );
  end generate;
end generate;

--
-- IF DPM generate (WIDTH/9) RAM BLOCKS with combined read/write ports
-- ie: use muxed write address for the read address also.
--

T: if (PMODE = 0) generate
   U: for i in 0 to (WIDTH/9) generate       
      V: RAM256x9SA
         port map( DI8 => wdata((i*9)+8),  DI7 => wdata((i*9)+7),
                   DI6 => wdata((i*9)+6),  DI5 => wdata((i*9)+5),
                   DI4 => wdata((i*9)+4),  DI3 => wdata((i*9)+3),
                   DI2 => wdata((i*9)+2),  DI1 => wdata((i*9)+1),
                   DI0 => wdata((i*9)+0),

                   WADDR7 => wadr(7),  WADDR6 => wadr(6),
                   WADDR5 => wadr(5),  WADDR4 => wadr(4),
                   WADDR3 => wadr(3),  WADDR2 => wadr(2),
                   WADDR1 => wadr(1),  WADDR0 => wadr(0),

                   WBLKB  => wrblk,    WRB => wrb,
                   WCLKS  => wr_2xclk,

                   WPE => open, RPE => open, DOS => open,

                   RBLKB  => GND,   RDB => GND,

                   RADDR7 => wadr(7),  RADDR6 => wadr(6),
                   RADDR5 => wadr(5),  RADDR4 => wadr(4),
                   RADDR3 => wadr(3),  RADDR2 => wadr(2),
                   RADDR1 => wadr(1),  RADDR0 => wadr(0),

                   DO8 => rd_data((i*9)+8), DO7 => rd_data((i*9)+7),
                   DO6 => rd_data((i*9)+6), DO5 => rd_data((i*9)+5),
                   DO4 => rd_data((i*9)+4), DO3 => rd_data((i*9)+3),
                   DO2 => rd_data((i*9)+2), DO1 => rd_data((i*9)+1),
                   DO0 => rd_data((i*9)+0),

                   PARODD => GND, DIS => GND
         );
  end generate;    
end generate;    
end RTL;


              